class Conversation:
    def __init__(self):
        self.conversation_history = []

    def add_message(self, role, content, function_name=None):
        # conditionally build message:
        if function_name is None:
            message = {"role": role, "content": content}
        else:
            message = {
                "role": role,
                "content": content,
                "name": function_name,
            }
        self.conversation_history.append(message)

    def display_conversation(self):
        for message in self.conversation_history:
            print(f"{message['role']}: {message['content']}\n\n")

    def display_last_message(self):
        last_message = self.conversation_history[-1]
        print(f"{last_message['role']}: {last_message['content']}\n\n")



from dataclasses import dataclass
from typing import Any, Callable, Dict


@dataclass
class FunctionSpec:
    func_name: str
    func_ref: Callable
    parameters: Dict[str, Any]

    @property
    def name(self) -> str:
        return self.func_name



from .function_spec import FunctionSpec
from .functions_orchestrator import FunctionsOrchestrator
from .metadata_generator import extract_openai_function_metadata, openai_function

__all__ = [
    "openai_function",
    "extract_openai_function_metadata",
    "FunctionsOrchestrator",
    "FunctionSpec",
]



"""Types module for openapi_functools """


def python_type_to_openapi_type(python_type: type) -> str:
    if python_type == str:
        return "string"
    elif python_type == int:
        return "integer"
    elif python_type == float:
        return "number"
    elif python_type == bool:
        return "boolean"
    elif python_type == list:
        return "array"
    elif python_type == dict:
        return "object"
    else:
        return "string"  # default



import json
from typing import Any, Callable, Dict, List, Optional

from openai_functools.function_spec import FunctionSpec
from openai_functools.metadata_generator import (
    construct_function_name,
    extract_openai_function_metadata,
)


class FunctionsOrchestrator:
    """
    Orchestrates the functions used in the OpenAI function calling models.
    """

    _functions: Dict[str, FunctionSpec]

    def __init__(self, functions: Optional[List[Callable]] = None) -> None:
        """
        Initializes the FunctionsOrchestrator with an optional list of functions.

        Args:
            functions (Optional[List[Callable]]): A list of functions to be registered.
        """
        self._functions = {}

        if functions is not None:
            for function in functions:
                self._add_function(function)

    @property
    def functions(self) -> List[Callable]:
        """
        Returns the list of registered functions.

        Returns:
            List[Callable]: The list of registered functions.
        """
        return self._functions

    @functions.setter
    def functions(self, functions: List[Callable]) -> None:
        """
        Sets the list of registered functions.

        Args:
            functions (List[Callable]): A list of functions to be registered.
        """
        for function in functions:
            self._add_function(function)

    @property
    def function_specs(self) -> List[FunctionSpec]:
        """
        Returns the list of function specifications for the registered functions.

        Returns:
            List[FunctionSpec]: The list of function specifications.
        """
        return self._functions.values()

    def register(self, function: Callable) -> None:
        """
        Registers a function.

        Args:
            function (Callable): The function to be registered.
        """
        self._add_function(function)

    def register_all(self, functions: List[Callable]) -> None:
        """
        Registers a list of functions.

        Args:
            functions (List[Callable]): The list of functions to be registered.
        """

        for function in functions:
            self._add_function(function)

    def register_instance(self, instance: Any) -> None:
        """
        Registers all methods of a single instance.

        Args:
            instance (Any): The instance whose methods are to be registered.
        """

        for method_name in dir(instance):
            method = getattr(instance, method_name)
            if not method_name.startswith("__") and callable(method):
                self._add_function(getattr(instance, method_name))

    def register_instances_all(self, instances: List[Any]):
        """
        Registers all methods of all instances.

        Args:
            instances (Any): The instances whose methods are to be registered.
        """

        for instance in instances:
            self.register_instance(instance)

    def _add_function(self, function: Callable) -> None:
        if not callable(function):
            raise TypeError(f'Function "{function}" is not callable.')

        function_name = construct_function_name(function)
        if function_name in self._functions:
            raise ValueError(f'Function "{function.__name__}" is already registered.')

        self._functions[function_name] = self._create_function_spec(function)

    def function(self, func: Optional[Callable] = None):
        """
        Registers a function if provided, otherwise returns a decorator for function registration.

        Args:
            func (Optional[Callable]): The function to be registered, if provided.

        Returns:
            Callable: The registered function or a decorator for function registration.
        """
        if func is not None:
            self.register(func)
            return func

        def wrapper(f):
            self.register(f)
            return f

        return wrapper

    def call_function(self, openai_response: dict) -> dict:
        """
        Calls a function based on the OpenAI response.

        Args:
            openai_response (dict): The OpenAI response containing the function call information.

        Returns:
            dict: The responses from the called function.
        """
        response_message = openai_response["choices"][0]["message"]

        if function_call := response_message.get("function_call"):
            function_name = function_call["name"]
            function_args = json.loads(function_call["arguments"])
            function = self._functions[function_name]

            if function is None:
                raise ValueError(
                    f'Function "{function_name}" is not '
                    f"registered with the orchestrator."
                )

            return function.func_ref(**function_args)
        else:
            raise ValueError(
                f'Function call information not found in response message "{response_message}".'
            )

    def _create_function_specs(self, functions: List[Callable]) -> List[FunctionSpec]:
        """
        Creates function specifications for a list of functions.

        Args:
            functions (List[Callable]): The list of functions for which to create specifications.

        Returns:
            List[FunctionSpec]: The list of created function specifications.
        """
        return [self._create_function_spec(function) for function in functions]

    @staticmethod
    def _create_function_spec(function: Callable) -> FunctionSpec:
        """
        Creates a function specification for a function.

        Args:
            function (Callable): The function for which to create a specification.

        Returns:
            FunctionSpec: The created function specification.
        """
        return FunctionSpec(
            func_name=construct_function_name(function),
            func_ref=function,
            parameters=extract_openai_function_metadata(function),
        )

    @property
    def function_descriptions(self) -> List[Dict[str, Any]]:
        """
        Returns the descriptions of the registered functions.

        Returns:
            List[Dict[str, Any]]: The list of function descriptions.
        """
        return [spec.parameters for spec in self._functions.values()]

    def create_function_descriptions(
        self, selected_functions: Optional[List[str]] = None
    ) -> List[Dict[str, Any]]:
        """
        Creates descriptions for the selected functions.

        Args:
            selected_functions (Optional[List[str]]): The list of selected function names. If None, descriptions for all registered functions are created.

        Returns:
            List[Dict[str, Any]]: The list of created function descriptions.
        """
        specs = (
            self._functions.values()
            if selected_functions is None
            else [
                spec
                for spec in self._functions.values()
                if spec.name in selected_functions
            ]
        )
        return [spec.parameters for spec in specs]



import inspect
import typing
from functools import wraps
from typing import Any, Callable

from docstring_parser import parse

from openai_functools.openai_types import python_type_to_openapi_type


def openai_function(func: Callable) -> Callable:
    """Wrapper for functions to add .openai_metadata property"""
    func.openai_metadata = extract_openai_function_metadata(func)

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return func(*args, **kwargs)

    return wrapper


def extract_openai_function_metadata(func: Callable) -> dict:
    """Extracts function metadata using function signature, docstring, ..."""
    sig = inspect.signature(func)
    params = sig.parameters
    properties = {}
    function_name = construct_function_name(func)

    # assumes the format from https://pypi.org/project/docstring-parser/
    docstring = parse(func.__doc__ if func.__doc__ else "")
    docstring_params = {param.arg_name: param.description for param in docstring.params}

    for name, param in params.items():
        properties[name] = extract_parameter_properties(param, docstring_params)

    metadata = {
        "name": function_name,
        "description": docstring.short_description or function_name,
        "parameters": {
            "type": "object",
            "properties": properties,
            "required": [
                name for name, param in params.items() if param.default == param.empty
            ],
        },
    }
    return metadata


# FIXME I believe this is in broken state, check py version thing? - Jakob 101023
def extract_literal_allowed_values(func):
    """Using signature to extract allowed values for Literal[...]"""
    signature = inspect.signature(func)
    literal_parameters = {}

    for param_name, param in signature.parameters.items():
        if isinstance(param.default, typing._LiteralGenericAlias):
            literal_parameters[param_name] = list(param.default.__args__)

    return literal_parameters


def extract_parameter_properties(
    param: inspect.Parameter, docstring_params: dict
) -> dict:
    """Extracts types of function parameters. Defaults to string if None found."""
    name = param.name

    properties = {
        "type": python_type_to_openapi_type(param.annotation)
        if param.annotation != param.empty
        else "string"  # make this configurable?
    }

    properties["description"] = docstring_params.get(name, name)

    if isinstance(param.default, typing._LiteralGenericAlias):
        properties["enum"] = list(param.default.__args__)
        properties["default"] = param.default.__args__[0]
    else:
        if param.default != param.empty:
            properties["default"] = param.default

    return properties


def construct_function_name(func: Callable) -> str:
    """Constructs a function name to uniquely identify a function or a method of an instance"""
    if not hasattr(func, "__self__"):
        return func.__name__
    else:
        # A hash is used to uniquely identify an instance
        return f"{func.__self__.__hash__()}__{func.__name__}"
